/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : PRC_WATCH, GPR_INITCLR

//`define USE_DMY_CLOCK
//`define PRC_WATCH
//`define GPR_INITCLR

// local defines : 

// GPR modele for the Reduced Core TinyAVR (attiny10) 
// version wk11
// 
// Target Device: xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal: Balanced
//  NORMAL     : wide reg. version (using gpr_hi and gpr_lo )
//  USE_GPR_16 : shrinked version ( small but slow )
//                                      
//                                   NORMAL    USE_GPR_16
// Number of Slice Flip Flops:           57      52
// Number of 4 input LUTs:              182     129
// Number of occupied Slices:           120      81
// Total Number of 4 input LUTs:        201     132
// Number used for Dual Port RAMs:       80      16
// Number of bonded IOBs:                82      79
// Post-PAR Static Timing Report
// Maximum frequency(MHz):           126.920  68.946
//
// Target Device: xc6slx9-2ftg256
//                                   NORMAL    USE_GPR_16
// Number of Slice Registers             52      48
// Number of Slice LUTs:                122      88
//  Number used as Dual Port RAM:        36       8
// Number of occupied Slices:            40      33
// Number of bonded IOBs:                78      79
// Post-PAR Static Timing Report
// Maximum frequency(MHz):           113.895  60.680
//
// Target Device: xc6slx9-3ftg256
//                                   NORMAL    USE_GPR_16
// Number of Slice Registers             53      48
// Number of Slice LUTs:                123      88
//  Number used as Dual Port RAM:        36       8
// Number of occupied Slices:            37      35
// Number of bonded IOBs:                78      79
// Post-PAR Static Timing Report
// (-3) Maximum frequency(MHz):      170.126  90.926

module rtavr_gpr_16(
		input RESET,
`ifdef SYS_DRIVEN_CLK4PH
		input CLK_0,
		input CLK_90,
		input CLK_180,
		input CLK_270,
`else // SYS_DRIVEN_CLK2X
		input CLK2X,
		input CLK,
`endif
		input WE,
		input PREDEC, POSTINC, 
		input [4:0] IDX_OFF,
		input [7:0] DI,
		input [2:0] ADDRAL31,
		input [3:0] ADDRAH,
		input [3:0] ADDRBL,
		input [3:0] ADDRBH,
		output [7:0] DOBL,
		output [7:0] DOBH
		, output [11:0] JA
		, output [15:0] INDEX
`ifdef PRC_WATCH
		, output [3:0] WB_ADDR
`endif
		);

`ifdef USE_GPR_16
// Timing MEMO:
//
//                T(N-1)               |            TN                 |
//              _______         _______         _______         _______
// CLK2X ______|       |_______|       |_______|       |_______|       |
//                       _______________                 _______________
// CLK _________________|               |_______________|               |
//
// ADDRAL,      ================================ (PM_DRIVEN_CLK2X) 
// ADDRBL                ================================ (else)     
// ADDRAH,ADDRBH,                       =================================
// PREDEC,POSTINC,WE
// DI                                                   ================
//                                     V                V      V
//                                 ADDRAL            ADDRAH   DI
//                                 ADDRBL            ADDRBH
//                   
// DOBL,DOBL                                                            |===
// INDEX(lo8)                                  ===========================..
// INDEX(hi8)                                           ==================
// JA(lo8)                                                      FFFFFFFFFF..
// JA(hi8)                                                      FFFFFFFFFF..

`ifdef SYS_DRIVEN_CLK4PH
    wire CLK_90_270 = (CLK_0 & CLK_90) | (CLK_180 & CLK_270);
    wire CLK_0_180  = (CLK_270 & CLK_0) | (CLK_90 & CLK_180);

    wire CLK = CLK_180;
    wire CLK_P = CLK_270;
    wire __CLK_0 = 1'b0;
    wire __CLK_90 =  1'b0;
    wire __CLK_180 = 1'b0;

`else // SYS_DRIVEN_CLK2X
    wire CLK_P = CLK;
    wire __CLK_0 = ~CLK;
    wire __CLK_90 = CLK;
    wire __CLK_180 = CLK;
`endif

  reg  [3:0] r_addra;
  reg  [3:0] r_addrb;

  reg  [7:0] s2_DOBL;
  reg  [7:0] s2_DOBH;
  reg  [7:0] gpr [0:15];

  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;

  reg  c;
  reg  ci;
  wire WB = (PREDEC | POSTINC);
  reg  [3:0] r_wb_addr;

  `ifdef PRC_WATCH
    assign WB_ADDR = r_wb_addr;
  `endif

  wire [7:0] DOA = gpr[r_addra];
  wire [7:0] DOB = gpr[r_addrb];

  wire [7:0] DOA2LP = DOA - 1;
  wire [8:0] DOA2LO = {1'b0 , DOA} + { 4'b0000 ,IDX_OFF };
  wire [7:0] DOA2H = (PREDEC &  c) ? DOB - 1 : DOB ;

  wire [8:0] WB_DATA = (PREDEC & (~CLK | c) ) ?  {1'b0 , DOA} - 1
                     : (POSTINC & (~CLK | c)) ?  {1'b0 , DOA} + 1
		     : {1'b0 , DOA};
  
  wire WE2 = WB | (WE & CLK) ;

  reg  [7:0] s1_DOAL;
  reg  [7:0] s1_DOBL;

  reg r_forward_to_bl;
  wire FORWARD_TO_BL = r_forward_to_bl;

  wire [7:0] v_DOBL = FORWARD_TO_BL ? DI : s1_DOBL;

  wire [7:0] v_DOBH = DOB;

    assign JA[7:0] = FORWARD_TO_BL ? DI : s1_DOBL;
    assign INDEX[7:0] = s1_DOAL;

    assign JA[11:8] = DOB[3:0];
	
    assign INDEX[15:13] = DOA2H[7:5];
    assign INDEX[12:8] =  DOA2H[4:0]
			+ ( ci ? 5'b00001 : 5'b00000 );

always @(`posedge_CLK_90_270)
  begin
      if (WE2) gpr[r_addra] <= (WB) ? WB_DATA[7:0] : DI ;
  end

always @(`posedge_CLK_90)
  begin
    if (RESET)
      begin
	s1_DOAL <= 0;
	r_wb_addr <= 0;
      end
    else if (~__CLK_90)
      begin
	s1_DOAL <= (PREDEC) ? DOA2LP[7:0] : DOA2LO[7:0] ;
	r_wb_addr <= (WE) ? ADDRAH : ADDRBH;
      end
  end

always @(`posedge_CLK_0_180)
  begin
    if (RESET)
      begin
	r_addra <= 0;
	r_addrb <= 0;
	c <= 0;
	ci <= 0;
      end
    else
      begin
	r_addra <= CLK_P ? {ADDRAL31 , 1'b0 }: r_wb_addr;
	r_addrb <= CLK_P ? ADDRBL : ADDRBH;
	c <= WB_DATA[8];
	ci <= DOA2LO[8];
      end
  end

always @(`posedge_CLK_0)
  begin
    if (RESET)
      begin
	s2_DOBL <= 0;
	s2_DOBH <= 0;
      end
    else if (~__CLK_0)
      begin
	s2_DOBL <= v_DOBL;
	s2_DOBH <= v_DOBH;
      end
  end

always @(`posedge_CLK_180)
  begin
    if (RESET)
      begin
	s1_DOBL <= 0;
	r_forward_to_bl <= 0;
      end
    else if (~__CLK_180)
      begin
	s1_DOBL <= DOB;
	r_forward_to_bl <=  (WE) & (r_wb_addr == r_addrb);
      end
  end

`ifdef GPR_INITCLR
  integer i;
  initial
    begin
      for (i = 0 ; i < 16 ; i = i + 1)
	gpr[i] <= 8'h00;
    end
`endif

`else // USE_GPR_16 
// Timing MEMO:
//
//                T(N-1)               |            TN                 |
//              _______         _______         _______         _______
// CLK2X ______|       |_______|       |_______|       |_______|       |
//                       _______________                 _______________
// CLK _________________|               |_______________|               |
//
// ADDRAL31,    ================================ (PM_DRIVEN_CLK2X) 
// ADDRBL                ================================ (else)     
// ADDRAH,ADDRBH,                       =================================
// PREDEC,POSTINC,WE
// DI                                                   ================
//                                     V                       V
//                                 ADDRAL                     DI
//                                 ADDRBL                  
//                   
// DOBL,DOBL                             
// INDEX(lo8)                           
// INDEX(hi8)                          
// JA(lo8)                            
// JA(hi8)                           
//
// READ OPs
//  
//  WE  PREDEC POSTiNC     DOBH(= Rr_in)      DOBL(= Rd_in)
//  0     -      -         gpr[ADDRBH]        gpr[ADDRBL] 
//  1     -      -         gpr[ADDRBH]/DI     gpr[ADDRBL]/DI 
//  
//  JA : ( non-latched { DOBH , DOBL }
//  
//  WE  PREDEC POSTINC     INDEX
//  0     0      0       { gpr[ADDRBH]    ,    gpr[ADDRAL]    }
//  0     1      0       { gpr[ADDRAH]    ,    gpr[ADDRAL]    } -1
//  0     0      1       { gpr[ADDRAH]    ,    gpr[ADDRAL]    }
//  
//  1     0      0       { gpr[ADDRBH]/DI ,    gpr[ADDRAL]/DI }
//  
// WRITE OPs
//  WE  PREDEC POSTINC     
//  0     1      0     --{ gpr[ADDRAH]    ,    gpr[ADDRAL]    } 
//  0     0      1       { gpr[ADDRAH]    ,    gpr[ADDRAL]    } ++
//  
//  1     0      0       gpr[prev ADDRAH] = DI
//  
`ifdef SYS_DRIVEN_CLK4PH
    wire __CLK_0 = 1'b0;
    wire __CLK_90 = 1'b0;
    wire __CLK_180 = 1'b0;
    wire __CLK_270 = 1'b0;
`else // SYS_DRIVEN_CLK2X
    wire __CLK_0 = ~CLK;
    wire __CLK_90 = CLK;
    wire __CLK_180 = CLK;
    wire __CLK_270 = ~CLK;
`endif

  reg  [2:0] r_addral;
  reg  r_addrbl_0;
  reg  [2:0] r_addrbl;
  reg  [2:0] r_wb_addr;
  reg  r_wb_addr_0;
  reg  [2:0] s2_addrbl;
  reg  s2_addrbl_0;
  reg  [2:0] s2_addrbh;
  reg  s2_addrbh_0;

  reg  [7:0] s2_DOBL;
  reg  [7:0] s2_DOBH;
  reg  [7:0] gpr_lo [0:7];
  reg  [7:0] gpr_hi [0:7];
  reg  [15:0] r_doa2;

  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;

  wire  [7:0] DOAL_AL = gpr_lo[r_addral];
  wire  [7:0] DOBL_BL = gpr_lo[s2_addrbl];
  wire  [7:0] DOBL_BH = gpr_lo[s2_addrbh];

  wire  [7:0] DOAH_WB = gpr_hi[r_wb_addr];
  wire  [7:0] DOBH_BL = gpr_hi[s2_addrbl];
  wire  [7:0] DOBH_BH = gpr_hi[s2_addrbh];

  wire WB = (PREDEC | POSTINC);

    wire [15:0] DOA = { DOAH_WB , DOAL_AL };


  `ifdef PRC_WATCH
    assign WB_ADDR = { r_wb_addr ,r_wb_addr_0 };
  `endif

  wire [15:0] DOA2 = (PREDEC) ? DOA - 1 : DOA;
  wire [15:0] WB_DATA = (PREDEC) ?  DOA - 1
                     : (POSTINC) ?  DOA + 1
		     : DOA;

  wire WE_AL = WB | (WE & ~r_wb_addr_0);
  wire WE_AH = WB | (WE &  r_wb_addr_0);

  wire [7:0] DI2_AL = (WB) ? WB_DATA[7:0] : DI;
  wire [7:0] DI2_AH = (WB) ? WB_DATA[15:8] : DI;

  wire [7:0] v_DOBL = s2_addrbl_0 ? DOBH_BL : DOBL_BL ;
  wire [7:0] v_DOBH = s2_addrbh_0 ? DOBH_BH : DOBL_BH ;

  assign JA[7:0] = DOBL_BL ;
  assign JA[11:8] = DOBH_BH[3:0];

  assign INDEX[15:13] = (WB) ? r_doa2[15:13] : DOBH_BH[7:5];
  assign INDEX[12:0] = { (WB) ? r_doa2[12:8] : DOBH_BH[4:0] 
			, r_doa2[7:0] }
		     + { 8'b00000000, IDX_OFF };

always @(`posedge_CLK_0) // negedge CLK
  begin
    if (RESET)
      begin
	r_addrbl_0 <= 0;
	r_addrbl   <= 0;
	s2_DOBL <= 0;
	s2_DOBH <= 0;
      end
    else if (~__CLK_0)
      begin
	r_addrbl_0 <= ADDRBL[0];
	r_addrbl   <= ADDRBL[3:1];
	s2_DOBL <= v_DOBL;
	s2_DOBH <= v_DOBH;
      end
  end

always @(`posedge_CLK_90)
  begin
    if (RESET)
      begin
	r_wb_addr <= 0;
	r_wb_addr_0 <= 0;
	r_addral <= 0;
      end
    else if (~__CLK_90)
      begin
	r_wb_addr_0 <=  (WE)? ADDRAH[0] : ADDRBH[0];
	r_wb_addr   <=  (WE)? ADDRAH[3:1] : ADDRBH[3:1];
`ifdef GPR_TUNE_1
	r_addral  <=  (~ADDRAH[0] & WE) ? ADDRAH[3:1] : ADDRAL31;
`else
	r_addral    <=  ADDRAL31;
`endif
      end
  end

always @(`posedge_CLK_180) // posedge CLK
  begin
    if (RESET)
      begin
	r_doa2 <= 0;
      end
    else if (~__CLK_180)
      begin
	r_doa2 <= DOA2;
      end
  end

always @(`posedge_CLK_270)
  begin
    if (RESET)
      begin
	s2_addrbl <= 0;
	s2_addrbl_0 <= 0;
	s2_addrbh <= 0;
	s2_addrbh_0 <= 0;
      end
    else if (~__CLK_270)
      begin
	if (WE_AH) gpr_hi[r_wb_addr] <= DI2_AH;
`ifdef GPR_TUNE_1
	if (WE_AL) gpr_lo[r_addral] <= DI2_AL;
`else
	if (WE_AL) gpr_lo[r_wb_addr] <= DI2_AL;
`endif
	s2_addrbl_0 <= r_addrbl_0;
	s2_addrbl   <= r_addrbl;
	s2_addrbh_0 <= ADDRBH[0];
	s2_addrbh   <= ADDRBH[3:1];
      end
  end

`ifdef GPR_INITCLR
  integer i;
  initial
    begin
      for (i = 0 ; i < 8 ; i = i + 1)
	begin
	  gpr_lo[i] <= 8'h00;
	  gpr_hi[i] <= 8'h00;
	end
    end
`endif

`endif // USE_GPR_16 
endmodule
